from typing import Any, Awaitable, Callable, Dict, Optional

from databases import Database
from graphql import graphql as graphql_exec
from nebulo.server.jwt import get_jwt_claims_handler
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import JSONResponse

__all__ = ["get_graphql_endpoint"]


def get_graphql_endpoint(
    schema: str, database: Database, jwt_secret: Optional[str]
) -> Callable[[Request], Awaitable[JSONResponse]]:
    """Retrieve the GraphQL variables from the Starlette Request"""

    get_jwt_claims = get_jwt_claims_handler(jwt_secret)

    async def graphql_endpoint(request):

        query = await get_query(request)
        variables = await get_variables(request)
        jwt_claims = await get_jwt_claims(request)
        request_context = {
            "request": request,
            "database": database,
            "query": query,
            "variables": variables,
            "jwt_claims": jwt_claims,
        }
        result = await graphql_exec(
            schema=schema, source=query, context_value=request_context, variable_values=variables
        )
        errors = result.errors
        result_dict = {"data": result.data, "errors": [error.formatted for error in errors or []]}
        return JSONResponse(result_dict)

    return graphql_endpoint


async def get_query(request: Request) -> Awaitable[str]:
    """Retrieve the GraphQL query from the Starlette Request"""

    content_type = request.headers.get("content-type", "")
    if content_type == "application/graphql":
        return await request.body()
    if content_type == "application/json":
        return (await request.json())["query"]
    raise HTTPException(400, "content-type header must be set")


async def get_variables(request) -> Awaitable[Dict[str, Any]]:
    """Retrieve the GraphQL variables from the Starlette Request"""

    content_type = request.headers.get("content-type", "")
    if content_type == "application/graphql":
        raise NotImplementedError("Getting variables from graphql content type not known")

    if content_type == "application/json":
        return (await request.json()).get("variables", {})
    raise HTTPException(400, "content-type header must be set")



import typing

from graphql.execution.execute import get_field_def
from graphql.execution.values import get_argument_values
from graphql.language import FieldNode, FragmentDefinitionNode, FragmentSpreadNode, InlineFragmentNode
from nebulo.gql.alias import Field, List, NonNull, ObjectType, ResolveInfo, Schema

__all__ = ["parse_resolve_info"]


def field_to_type(field):
    """Recursively unwraps nested Field, List, and NonNull
    qualifiers to a concrete GraphQL Type"""
    if isinstance(field, Field):
        return field_to_type(field.type)
    if isinstance(field, List):
        return field_to_type(field.of_type)
    if isinstance(field, NonNull):
        return field_to_type(field.of_type)
    return field


class ASTNode:
    def __init__(
        self,
        field_node: FieldNode,
        field_def: ObjectType,
        schema: Schema,
        parent: typing.Optional["ASTNode"],
        variable_values,
        parent_type,
        fragments: typing.Dict[str, FragmentDefinitionNode],
    ):
        self.name = field_node.name.value

        # A connection/edge/etc class
        field_def = get_field_def(schema, parent_type, self.name)

        _args = get_argument_values(type_def=field_def, node=field_node, variable_values=variable_values)

        selection_set = field_node.selection_set
        field_type = field_to_type(field_def)

        self.alias = (field_node.alias.value if field_node.alias else None) or field_node.name.value
        self.return_type = field_type
        self.parent: typing.Optional[ASTNode] = parent
        self.parent_type = parent_type
        self.args: typing.Dict[str, typing.Any] = _args
        self.path: typing.List[str] = parent.path + [self.name] if parent is not None else ["root"]

        def from_selection_set(selection_set):
            for selection_ast in selection_set.selections:

                # Handle fragments
                if isinstance(selection_ast, FragmentSpreadNode):
                    fragment_name = selection_ast.name.value
                    fragment = fragments[fragment_name]
                    fragment_selection_set = fragment.selection_set
                    yield from from_selection_set(fragment_selection_set)

                elif isinstance(selection_ast, InlineFragmentNode):
                    yield from from_selection_set(selection_ast.selection_set)

                else:
                    selection_name = selection_ast.name.value
                    if selection_name.startswith("__"):
                        # Reserved "introspection" field handled by framework
                        continue
                    selection_field = field_type.fields[selection_name]
                    yield ASTNode(
                        field_node=selection_ast,
                        field_def=selection_field,
                        schema=schema,
                        parent=self,
                        variable_values=variable_values,
                        parent_type=field_type,
                        fragments=fragments,
                    )

        sub_fields = list(from_selection_set(selection_set)) if selection_set else []
        self.fields = sub_fields

    def get_subfield_alias(self, path: typing.List[str]):
        """Searches subfields by name returning the alias
        for the terminating element in *path* list. If that subfield
        is not in the selection set, return the terminating elements name"""
        if len(path) < 1:
            raise Exception("Path must contain an element")

        for subfield in self.fields:
            if subfield.name == path[0]:
                if len(path) == 1:
                    return subfield.alias or path[0]
                return subfield.get_subfield_alias(path[1:])
        return path[-1]


def parse_resolve_info(info: ResolveInfo) -> ASTNode:
    """Converts execution ResolveInfo into a dictionary
    hierarchy

    {
        "alias": *alias*,
        "name": *name*,
        "return_type": *return_type*,
        "args":  {
            "first": 10
        },
        "parent": <reference to parent field or None>
        "fields": [
            # Same structure again, for each field selected
            # in the query
            {
                "alias": ...
                "name": ...
            }
        }
    }
    """
    # Root info
    field_node = info.field_nodes[0]
    schema = info.schema

    fragments: typing.Dict[str, FragmentDefinitionNode] = info.fragments

    # Current field from parent
    parent_type = info.parent_type
    parent_lookup_name = field_node.name.value
    current_field = parent_type.fields[parent_lookup_name]
    parsed_info = ASTNode(
        field_node,
        current_field,
        schema,
        parent=None,
        variable_values=info.variable_values,
        parent_type=info.parent_type,
        fragments=fragments,
    )
    return parsed_info



import typing

from graphql.pyutils import Path
from nebulo.gql.alias import ResolveInfo


def default_resolver(_, info: ResolveInfo, **kwargs) -> typing.Any:
    """Expects the final, complete result to exist in context['result']
    and uses the current path to retrieve and return the expected result for
    the current location in the query

    Why TF would you do this?
    - To avoid writing a custom executor
    - To keep the resolver tree extensible for end users
    - To keep everything as default as possible
    """
    path: Path = info.path
    context = info.context
    final_result = context["result"]

    remain_result = final_result
    for elem in path.as_list():
        remain_result = remain_result[elem]

    return remain_result



# pylint: disable=invalid-name
from __future__ import annotations

from nebulo.config import Config
from nebulo.gql.alias import CreatePayloadType, DeletePayloadType, UpdatePayloadType
from nebulo.gql.parse_info import ASTNode
from nebulo.gql.relay.node_interface import to_node_id_sql
from nebulo.gql.resolve.transpile.query_builder import field_name_to_column
from nebulo.sql.inspect import get_primary_key_columns


def build_mutation(tree: ASTNode):
    """Dispatch for Mutation Types

    Returns an executable sqlalchemy statment and a result post-processor
    """

    if isinstance(tree.return_type, CreatePayloadType):
        return build_insert(tree)
    elif isinstance(tree.return_type, UpdatePayloadType):
        return build_update(tree)
    elif isinstance(tree.return_type, DeletePayloadType):
        return build_delete(tree)
    else:
        raise Exception("Unknown mutation type")


def build_insert(tree: ASTNode):
    return_type = tree.return_type
    # Find the table type
    return_sqla_model = return_type.sqla_model
    table_input_arg_name = Config.table_name_mapper(return_sqla_model)
    input_values = tree.args["input"][table_input_arg_name]
    col_name_to_value = {}
    for arg_name, arg_value in input_values.items():
        col = field_name_to_column(return_sqla_model, arg_name)
        col_name_to_value[col.name] = arg_value

    core_table = return_sqla_model.__table__

    query = (
        core_table.insert()
        .values(**col_name_to_value)
        .returning(to_node_id_sql(return_sqla_model, core_table).label("nodeId"))
    )
    return query


def build_update(tree: ASTNode):
    return_type = tree.return_type
    sqla_model = return_type.sqla_model

    # Where Clause
    pkey_cols = get_primary_key_columns(sqla_model)
    node_id = tree.args["input"]["nodeId"]
    pkey_clause = [col == node_id.values[str(col.name)] for col in pkey_cols]

    # Find the table type
    return_sqla_model = return_type.sqla_model
    table_input_arg_name = Config.table_name_mapper(return_sqla_model)
    input_values = tree.args["input"][table_input_arg_name]
    col_name_to_value = {}
    for arg_name, arg_value in input_values.items():
        col = field_name_to_column(return_sqla_model, arg_name)
        col_name_to_value[col.name] = arg_value

    core_table = return_sqla_model.__table__

    query = (
        core_table.update()
        .where(*pkey_clause)
        .values(**col_name_to_value)
        .returning(to_node_id_sql(return_sqla_model, core_table).label("nodeId"))
    )
    return query


def build_delete(tree: ASTNode):
    return_type = tree.return_type
    sqla_model = return_type.sqla_model

    # Where Clause
    pkey_cols = get_primary_key_columns(sqla_model)
    node_id = tree.args["input"]["nodeId"]
    pkey_clause = [col == node_id.values[str(col.name)] for col in pkey_cols]

    # Find the table type
    return_sqla_model = return_type.sqla_model
    core_table = return_sqla_model.__table__

    query = (
        core_table.delete().where(*pkey_clause).returning(to_node_id_sql(return_sqla_model, core_table).label("nodeId"))
    )
    return query



from __future__ import annotations

from functools import lru_cache

from nebulo.config import Config
from nebulo.gql.alias import (
    Field,
    InputObjectType,
    NonNull,
    ObjectType,
    String,
    TableInputType,
    UpdateInputType,
    UpdatePayloadType,
)
from nebulo.gql.convert.column import convert_column_to_input
from nebulo.gql.relay.node_interface import ID
from nebulo.gql.resolve.resolvers.default import default_resolver
from nebulo.sql.inspect import get_columns
from nebulo.sql.table_base import TableProtocol

"""
updateAccount(input: UpdateAccountInput!):
    UpdateAccountPayload
"""


@lru_cache()
def update_entrypoint_factory(sqla_model: TableProtocol, resolver) -> Field:
    """updateAccount"""
    relevant_type_name = Config.table_type_name_mapper(sqla_model)
    name = f"update{relevant_type_name}"
    args = {"input": NonNull(update_input_type_factory(sqla_model))}
    payload = update_payload_factory(sqla_model)
    return {
        name: Field(
            payload,
            args=args,
            resolve=resolver,
            description=f"Updates a single {relevant_type_name} using its globally unique id and a patch.",
        )
    }


@lru_cache()
def update_input_type_factory(sqla_model: TableProtocol) -> InputObjectType:
    """UpdateAccountInput!"""
    relevant_type_name = Config.table_type_name_mapper(sqla_model)
    result_name = f"Update{relevant_type_name}Input"

    input_object_name = Config.table_name_mapper(sqla_model)

    attrs = {
        "nodeId": NonNull(ID),
        "clientMutationId": String,
        input_object_name: NonNull(patch_type_factory(sqla_model)),
    }
    return UpdateInputType(result_name, attrs, description=f"All input for the create {relevant_type_name} mutation.")


def patch_type_factory(sqla_model: TableProtocol) -> InputObjectType:
    """AccountPatch"""
    relevant_type_name = Config.table_type_name_mapper(sqla_model)
    result_name = f"{relevant_type_name}Patch"

    attrs = {}
    for column in get_columns(sqla_model):
        if not Config.exclude_update(column):
            field_key = Config.column_name_mapper(column)
            column_field = convert_column_to_input(column)
            # TODO Unwrap not null here
            attrs[field_key] = column_field
    return TableInputType(result_name, attrs, description=f"An input for mutations affecting {relevant_type_name}.")


@lru_cache()
def update_payload_factory(sqla_model: TableProtocol) -> InputObjectType:
    """UpdateAccountPayload"""
    from nebulo.gql.convert.table import table_factory

    relevant_type_name = Config.table_type_name_mapper(sqla_model)
    relevant_attr_name = Config.table_name_mapper(sqla_model)
    result_name = f"Update{relevant_type_name}Payload"

    attrs = {
        "clientMutationId": Field(String, resolve=default_resolver),
        "nodeId": ID,
        relevant_attr_name: Field(
            table_factory(sqla_model),
            resolve=default_resolver,
            description=f"The {relevant_type_name} that was created by this mutation.",
        ),
    }

    return UpdatePayloadType(
        result_name, attrs, description=f"The output of our update {relevant_type_name} mutation", sqla_model=sqla_model
    )



from __future__ import annotations

import click
import uvicorn
from graphql.utilities import print_schema
from nebulo.env import EnvManager
from sqlalchemy import create_engine


@click.group()
@click.version_option(version="0.0.1")
def main(**kwargs):
    pass  # pragma: no cover


@main.command()
@click.option("-c", "--connection", default="sqlite:///")
@click.option("-p", "--port", default=5034)
@click.option("-h", "--host", default="0.0.0.0")
@click.option("-w", "--workers", default=1)
@click.option("-s", "--schema", default="public")
@click.option("--jwt-identifier", default=None)
@click.option("--jwt-secret", default=None)
@click.option("--reload/--no-reload", default=False)
def run(connection, schema, host, port, jwt_identifier, jwt_secret, reload, workers):
    """Run the GraphQL Server"""
    if reload and workers > 1:
        print("Reload not supported with workers > 1")
    else:

        with EnvManager(
            NEBULO_CONNECTION=connection,
            NEBULO_SCHEMA=schema,
            NEBULO_JWT_IDENTIFIER=jwt_identifier,
            NEBULO_JWT_SECRET=jwt_secret,
        ):

            uvicorn.run("nebulo.server.app:APP", host=host, workers=workers, port=port, log_level="info", reload=reload)


@main.command()
@click.option("-c", "--connection", default="sqlite:///")
@click.option("-s", "--schema", default="public")
@click.option("-o", "--out-file", type=click.File("w"), default=None)
def dump_schema(connection, schema, out_file):
    """Dump the GraphQL Schema to stdout or file"""
    from nebulo.gql.sqla_to_gql import sqla_models_to_graphql_schema
    from nebulo.sql.reflection.manager import reflect_sqla_models

    engine = create_engine(connection)
    sqla_models, sql_functions = reflect_sqla_models(engine, schema=schema)
    schema = sqla_models_to_graphql_schema(sqla_models, sql_functions)
    schema_str = print_schema(schema)
    click.echo(schema_str, file=out_file)



"""
A base class to derive sql tables from
"""

from sqlalchemy import Table
from typing_extensions import Protocol, runtime_checkable


@runtime_checkable
class TableProtocol(Protocol):
    __table__: Table



from pygments.lexer import RegexLexer
from pygments.token import Comment, Keyword, Name, Number, Operator, Punctuation, String, Text

__all__ = ["GraphQLLexer"]


class GraphQLLexer(RegexLexer):
    """
    Pygments GraphQL lexer for mkdocs
    """

    name = "GraphQL"
    aliases = ["graphql", "gql"]
    filenames = ["*.graphql", "*.gql"]
    mimetypes = ["application/graphql"]

    tokens = {
        "root": [
            (r"#.*", Comment.Singline),
            (r'"""("?"?(\\"""|\\(?!=""")|[^"\\]))*"""', Comment.Multi),
            (r"\.\.\.", Operator),
            # u
            (r'".*"', String.Double),
            (r"(-?0|-?[1-9][0-9]*)(\.[0-9]+[eE][+-]?[0-9]+|\.[0-9]+|[eE][+-]?[0-9]+)", Number.Float),
            (r"(-?0|-?[1-9][0-9]*)", Number.Integer),
            (r"\$+[_A-Za-z][_0-9A-Za-z]*", Name.Variable),
            (r"[_A-Za-z][_0-9A-Za-z]+\s?:", Text),
            (
                r"(type|query|interface|mutation|extend|input|implements|directive|@[a-z]+|on|true|false|null)\b",
                Keyword.Type,
            ),
            (r"[!$():=@\[\]{|}]+?", Punctuation),
            (r"[_A-Za-z][_0-9A-Za-z]*", Keyword),
            (r"(\s|,)", Text),
        ]
    }



# pylint: disable=comparison-with-callable
from __future__ import annotations

import json
import typing

from nebulo.gql.alias import ScalarType
from nebulo.sql.inspect import get_primary_key_columns, get_table_name
from nebulo.sql.statement_helpers import literal_string
from nebulo.text_utils.base64 import from_base64, to_base64
from sqlalchemy import func
from sqlalchemy.sql.selectable import Alias


class CursorStructure(typing.NamedTuple):
    table_name: str
    values: typing.Dict[str, typing.Any]

    @classmethod
    def from_dict(cls, contents: typing.Dict) -> CursorStructure:
        res = cls(table_name=contents["table_name"], values=contents["values"])
        return res

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {"table_name": self.table_name, "values": self.values}

    def serialize(self) -> str:
        ser = to_base64(json.dumps(self.to_dict()))
        return ser

    @classmethod
    def deserialize(cls, serialized: str) -> CursorStructure:
        contents = json.loads(from_base64(serialized))
        return cls.from_dict(contents)


def serialize(value: typing.Union[CursorStructure, typing.Dict]):
    node_id = CursorStructure.from_dict(value) if isinstance(value, dict) else value
    return node_id.serialize()


def to_cursor_sql(sqla_model, query_elem: Alias):
    table_name = get_table_name(sqla_model)

    pkey_cols = get_primary_key_columns(sqla_model)

    # Columns selected from query element
    vals = []
    for col in pkey_cols:
        col_name = str(col.name)
        vals.extend([literal_string(col_name), query_elem.c[col_name]])

    return func.jsonb_build_object(
        literal_string("table_name"),
        literal_string(table_name),
        literal_string("values"),
        func.jsonb_build_object(*vals),
    )


Cursor = ScalarType(
    "Cursor",
    description="Pagination point",
    serialize=serialize,
    parse_value=CursorStructure.deserialize,
    parse_literal=lambda x: CursorStructure.deserialize(x.value),
)



# pylint: disable=unused-argument
from __future__ import annotations

from nebulo.text_utils import snake_to_camel, to_plural
from sqlalchemy import Table


def rename_table(base, tablename: str, table: Table) -> str:
    """Produce a 'camelized' class name, e.g. 'words_and_underscores' -> 'WordsAndUnderscores'"""
    return snake_to_camel(tablename, upper=True)


def rename_to_one_collection(base, local_cls, referred_cls, constraint) -> str:
    referred_name = referred_cls.__name__
    return referred_name + "_by" + "_and".join(col.name for col in constraint.columns)


def rename_to_many_collection(base, local_cls, referred_cls, constraint) -> str:
    "Produce an 'uncamelized', 'pluralized' class name, e.g. "
    "'SomeTerm' -> 'some_terms'"
    referred_name = referred_cls.__name__
    pluralized = to_plural(referred_name)
    return pluralized + "_by" + "_and".join(col.name for col in constraint.columns)



from __future__ import annotations

from functools import lru_cache

from nebulo.config import Config
from nebulo.gql.alias import (
    CreateInputType,
    CreatePayloadType,
    Field,
    InputObjectType,
    NonNull,
    ObjectType,
    String,
    TableInputType,
)
from nebulo.gql.convert.column import convert_column_to_input
from nebulo.gql.relay.node_interface import ID
from nebulo.gql.resolve.resolvers.default import default_resolver
from nebulo.sql.inspect import get_columns
from nebulo.sql.table_base import TableProtocol

"""
createAccount(input: CreateAccountInput!):
    CreateAccountPayload
"""


@lru_cache()
def create_entrypoint_factory(sqla_model: TableProtocol, resolver) -> Field:
    """createAccount"""
    relevant_type_name = Config.table_type_name_mapper(sqla_model)
    name = f"create{relevant_type_name}"
    args = {"input": NonNull(create_input_type_factory(sqla_model))}
    payload = create_payload_factory(sqla_model)
    return {name: Field(payload, args=args, resolve=resolver, description=f"Creates a single {relevant_type_name}.")}


@lru_cache()
def create_input_type_factory(sqla_model: TableProtocol) -> CreateInputType:
    """CreateAccountInput!"""
    relevant_type_name = Config.table_type_name_mapper(sqla_model)
    result_name = f"Create{relevant_type_name}Input"

    input_object_name = Config.table_name_mapper(sqla_model)

    attrs = {"clientMutationId": String, input_object_name: NonNull(input_type_factory(sqla_model))}
    return CreateInputType(result_name, attrs, description=f"All input for the create {relevant_type_name} mutation.")


def input_type_factory(sqla_model: TableProtocol) -> TableInputType:
    """AccountInput"""
    relevant_type_name = Config.table_type_name_mapper(sqla_model)
    result_name = f"{relevant_type_name}Input"

    attrs = {}
    for column in get_columns(sqla_model):
        if not Config.exclude_insert(column):
            field_key = Config.column_name_mapper(column)
            attrs[field_key] = convert_column_to_input(column)
    return TableInputType(result_name, attrs, description=f"An input for mutations affecting {relevant_type_name}.")


@lru_cache()
def create_payload_factory(sqla_model: TableProtocol) -> CreatePayloadType:
    """CreateAccountPayload"""
    from nebulo.gql.convert.table import table_factory

    relevant_type_name = Config.table_type_name_mapper(sqla_model)
    relevant_attr_name = Config.table_name_mapper(sqla_model)
    result_name = f"Create{relevant_type_name}Payload"

    attrs = {
        "clientMutationId": Field(String, resolve=default_resolver),
        "nodeId": ID,
        relevant_attr_name: Field(
            NonNull(table_factory(sqla_model)),
            description=f"The {relevant_type_name} that was created by this mutation.",
            resolve=default_resolver,
        ),
    }

    return CreatePayloadType(
        result_name, attrs, description=f"The output of our create {relevant_type_name} mutation", sqla_model=sqla_model
    )



# pylint: disable=invalid-name,unsubscriptable-object
from __future__ import annotations

from functools import lru_cache

from nebulo.config import Config
from nebulo.gql.alias import Argument, Field, NonNull, TableType
from nebulo.gql.convert.column import convert_column
from nebulo.gql.relay.node_interface import ID, NodeInterface
from nebulo.gql.resolve.resolvers.default import default_resolver
from nebulo.sql.inspect import get_columns, get_relationships, is_nullable
from nebulo.sql.table_base import TableProtocol
from sqlalchemy.orm import interfaces


def table_field_factory(sqla_model: TableProtocol, resolver) -> Field:
    relevant_type_name = Config.table_type_name_mapper(sqla_model)
    node = table_factory(sqla_model)
    return Field(
        node,
        args={"nodeId": Argument(NonNull(ID))},
        resolve=resolver,
        description=f"Reads a single {relevant_type_name} using its globally unique ID",
    )


@lru_cache()
def table_factory(sqla_model: TableProtocol) -> TableType:
    """
    Reflects a SQLAlchemy table into a graphql-core GraphQLObjectType

    Parameters
    ----------
    sqla_model
        A SQLAlchemy ORM Table

    """
    from .connection import connection_field_factory

    name = Config.table_type_name_mapper(sqla_model)

    def build_attrs():
        attrs = {}

        # Override id to relay standard
        attrs["nodeId"] = Field(NonNull(ID), resolve=default_resolver)

        for column in get_columns(sqla_model):
            if not Config.exclude_read(column):
                key = Config.column_name_mapper(column)
                attrs[key] = convert_column(column)

        for relationship in get_relationships(sqla_model):
            direction = relationship.direction
            to_sqla_model = relationship.mapper.class_
            relationship_is_nullable = is_nullable(relationship)

            # Name of the attribute on the model
            attr_key = Config.relationship_name_mapper(relationship)

            # If this model has 1 counterpart, do not use a list
            if direction == interfaces.MANYTOONE:
                _type = table_factory(to_sqla_model)
                _type = NonNull(_type) if not relationship_is_nullable else _type
                attrs[attr_key] = Field(_type, resolve=default_resolver)

            # Otherwise, set it up as a connection
            elif direction in (interfaces.ONETOMANY, interfaces.MANYTOMANY):
                connection_field = connection_field_factory(
                    to_sqla_model, resolver=default_resolver, not_null=relationship_is_nullable
                )
                attrs[attr_key] = connection_field

        return attrs

    return_type = TableType(
        name=name, fields=build_attrs, interfaces=[NodeInterface], description="", sqla_model=sqla_model
    )

    return return_type



from __future__ import annotations

import json
import typing

from flupy import flu
from nebulo.config import Config
from nebulo.gql.alias import (
    CompositeType,
    ConnectionType,
    CreatePayloadType,
    DeletePayloadType,
    FunctionPayloadType,
    MutationPayloadType,
    ObjectType,
    ResolveInfo,
    ScalarType,
    TableType,
    UpdatePayloadType,
)
from nebulo.gql.parse_info import parse_resolve_info
from nebulo.gql.relay.node_interface import NodeIdStructure, to_node_id_sql
from nebulo.gql.resolve.transpile.mutation_builder import build_mutation
from nebulo.gql.resolve.transpile.query_builder import sql_builder, sql_finalize
from nebulo.sql.table_base import TableProtocol
from sqlalchemy import Text, func, literal, select


async def async_resolver(_, info: ResolveInfo, **kwargs) -> typing.Any:
    """Awaitable GraphQL Entrypoint resolver

    Expects:
        info.context['database'] to contain a databases.Database
    """
    context = info.context
    database = context["database"]
    jwt_claims = context["jwt_claims"]
    tree = parse_resolve_info(info)

    async with database.transaction():

        if jwt_claims:
            # Setting local variables an not be done in prepared statement
            # since JWT claims are signed, literal binds should be ok
            claims = [
                func.set_config(
                    literal("jwt.claims.").op("||")(func.cast(claim_key, Text())),
                    func.cast(str(claim_value), Text()),
                    True,
                )
                for claim_key, claim_value in jwt_claims.items()
            ]
            await database.execute(select(claims))

        result: typing.Dict[str, typing.Any]

        if isinstance(tree.return_type, FunctionPayloadType):
            sql_function = tree.return_type.sql_function
            function_args = [val for key, val in tree.args["input"].items() if key != "clientMutationId"]
            func_call = sql_function.to_executable(function_args)
            stmt = select([func_call.label("result")])

            # Function returning table row
            if isinstance(sql_function.return_sqla_type, TableProtocol):
                return_sqla_model = sql_function.return_sqla_type
                core_table = return_sqla_model.__table__
                stmt = select([to_node_id_sql(return_sqla_model, core_table.alias()).label("nodeId")]).select_from(
                    stmt.alias()
                )
                stmt_result = await database.fetch_one(query=stmt)
                row = json.loads(stmt_result["nodeId"])
                node_id = NodeIdStructure.from_dict(row)

                # Add nodeId to AST and query
                query_tree = next(iter([x for x in tree.fields if x.name == "result"]), None)
                if query_tree is not None:
                    query_tree.args["nodeId"] = node_id
                    base_query = sql_builder(query_tree)
                    query = sql_finalize(query_tree.alias, base_query)
                    coro_rvf_result: str = (await database.fetch_one(query=query))["json"]
                    stmt_result = json.loads(coro_rvf_result)
                else:
                    stmt_result = {}
            else:
                stmt_result = await database.fetch_one(query=stmt)

            maybe_mutation_id = tree.args["input"].get("clientMutationId")
            mutation_id_alias = next(
                iter([x.alias for x in tree.fields if x.name == "clientMutationId"]),
                "clientMutationId",
            )
            result = {tree.alias: {**stmt_result, **{mutation_id_alias: maybe_mutation_id}}}

        elif isinstance(tree.return_type, MutationPayloadType):
            stmt = build_mutation(tree)
            stmt_result = await database.fetch_one(query=stmt)
            row = json.loads(stmt_result["nodeId"])
            node_id = NodeIdStructure.from_dict(row)

            maybe_mutation_id = tree.args["input"].get("clientMutationId")
            mutation_id_alias = next(
                iter([x.alias for x in tree.fields if x.name == "clientMutationId"]),
                "clientMutationId",
            )
            node_id_alias = next(iter([x.alias for x in tree.fields if x.name == "nodeId"]), "nodeId")
            output_row_name: str = Config.table_name_mapper(tree.return_type.sqla_model)
            query_tree = next(iter([x for x in tree.fields if x.name == output_row_name]), None)
            sql_result = {}
            if query_tree:
                # Set the nodeid of the newly created record as an arg
                query_tree.args["nodeId"] = node_id
                base_query = sql_builder(query_tree)
                query = sql_finalize(query_tree.alias, base_query)
                coro_result: str = (await database.fetch_one(query=query))["json"]
                sql_result = json.loads(coro_result)
            result = {
                tree.alias: {**sql_result, mutation_id_alias: maybe_mutation_id},
                mutation_id_alias: maybe_mutation_id,
                node_id_alias: node_id,
            }

        elif isinstance(tree.return_type, (ObjectType, ScalarType)):
            base_query = sql_builder(tree)
            query = sql_finalize(tree.name, base_query)

            query_coro = database.fetch_one(query=query)
            coro_result = await query_coro
            str_result: str = coro_result["json"]  # type: ignore

            query_json_result = json.loads(str_result)

            if isinstance(tree.return_type, ScalarType):
                # If its a scalar, unwrap the top level name
                result = flu(query_json_result.values()).first(None)
            else:
                result = query_json_result

        else:
            raise Exception("sql builder could not handle return type")

    # Stash result on context to enable dumb resolvers to not fail
    context["result"] = result
    return result



from inspect import isclass
from typing import Union

from nebulo.env import EnvManager
from nebulo.sql.inspect import get_comment, get_table_name
from nebulo.sql.reflection.function import SQLFunction
from nebulo.sql.reflection.views import ViewMixin
from nebulo.sql.table_base import TableProtocol
from nebulo.text_utils import snake_to_camel, to_plural
from sqlalchemy.orm import RelationshipProperty
from sqlalchemy.sql.schema import Column
from typing_extensions import Literal

ENV = EnvManager.get_environ()


class Config:

    CONNECTION = ENV.get("NEBULO_CONNECTION")
    SCHEMA = ENV.get("NEBULO_SCHEMA")
    JWT_IDENTIFIER = ENV.get("NEBULO_JWT_IDENTIFIER")
    JWT_SECRET = ENV.get("NEBULO_JWT_SECRET")

    @staticmethod
    def function_name_mapper(sql_function: SQLFunction) -> str:
        """to_upper -> toUpper"""
        return snake_to_camel(sql_function.name, upper=False)

    @staticmethod
    def function_type_name_mapper(sql_function: SQLFunction) -> str:
        """to_upper -> ToUpper"""
        return snake_to_camel(sql_function.name, upper=True)

    @staticmethod
    def table_type_name_mapper(sqla_table: TableProtocol) -> str:
        table_name = get_table_name(sqla_table)
        return snake_to_camel(table_name)

    @classmethod
    def table_name_mapper(cls, sqla_table: TableProtocol) -> str:
        """Return the type name with the first character lower case"""
        type_name = cls.table_type_name_mapper(sqla_table)
        return type_name[0].lower() + type_name[1:]

    @staticmethod
    def column_name_mapper(column: Column) -> str:
        return snake_to_camel(column.name, upper=False)

    @staticmethod
    def relationship_name_mapper(relationship: RelationshipProperty) -> str:
        # Union of Mapper or ORM instance
        referred_cls = relationship.argument
        if hasattr(referred_cls, "class_"):
            referred_cls = referred_cls.class_

        referred_name = get_table_name(referred_cls)
        cardinal_name = to_plural(referred_name) if relationship.uselist else referred_name
        camel_name = snake_to_camel(cardinal_name, upper=False)
        relationship_name = (
            camel_name
            + "By"
            + "And".join(
                snake_to_camel(local_col.name, upper=True) + "To" + snake_to_camel(remote_col.name, upper=True)
                for local_col, remote_col in relationship.local_remote_pairs
            )
        )
        return relationship_name

    # SQL Comment Config
    @staticmethod
    def _exclude_check(
        entity: Union[TableProtocol, Column],
        operation: Union[Literal["read"], Literal["insert"], Literal["update"], Literal["delete"]],
    ) -> bool:
        """Shared SQL comment parsing logic for excludes"""
        comment: str = get_comment(entity)
        lines = comment.split("\n")
        for line in lines:
            if "@exclude" in line and operation in line:
                return True
        return False

    @classmethod
    def exclude_read(cls, entity: Union[TableProtocol, Column]) -> bool:
        """Should the entity be excluded from reads?"""
        return cls._exclude_check(entity, "read")

    @classmethod
    def exclude_insert(cls, entity: Union[TableProtocol, Column]) -> bool:
        """Should the entity be excluded from inserts?"""
        # Views do not support insert
        if isclass(entity) and issubclass(entity, ViewMixin):  # type: ignore
            return True

        return cls._exclude_check(entity, "insert")

    @classmethod
    def exclude_update(cls, entity: Union[TableProtocol, Column]) -> bool:
        """Should the entity be excluded from updates?"""

        # Views do not support updates
        if isclass(entity) and issubclass(entity, ViewMixin):  # type: ignore
            return True

        return cls._exclude_check(entity, "update")

    @classmethod
    def exclude_delete(cls, entity: Union[TableProtocol, Column]) -> bool:
        """Should the entity be excluded from deletes?"""

        # Views do not support deletes
        if isclass(entity) and issubclass(entity, ViewMixin):  # type: ignore
            return True

        return cls._exclude_check(entity, "delete")



from nebulo.config import Config
from nebulo.server.starlette import create_app

APP = create_app(
    schema=Config.SCHEMA,
    connection=Config.CONNECTION,
    jwt_identifier=Config.JWT_IDENTIFIER,
    jwt_secret=Config.JWT_SECRET,
)



# pylint: disable=unused-argument,invalid-name,line-too-long,unsubscriptable-object
from __future__ import annotations

from typing import List

from nebulo.sql.reflection.names import rename_table
from nebulo.sql.table_base import TableProtocol
from sqlalchemy import ForeignKeyConstraint, PrimaryKeyConstraint, Table
from sqlalchemy import text as sql_text


class ViewMixin:
    """Mixin to differentiate views from tables"""

    is_view = True


def reflect_views(engine, schema, declarative_base) -> List[TableProtocol]:
    """Reflect SQLAlchemy ORM Tables from the database"""

    sql = sql_text(
        """
    select
	    relname view_name, description view_comment
    from
	    pg_views v
	    left join pg_description c
		    on true
	    left join pg_class on c.objoid = pg_class.oid
	    left join pg_namespace on pg_class.relnamespace = pg_namespace.oid
    where
	    v.viewname = relname
	    and nspname= :schema
    """
    )
    rows = engine.execute(sql, schema=schema).fetchall()

    views: List[TableProtocol] = []

    for view_name, view_comment in rows:
        primary_key_constraint = reflect_virtual_primary_key_constraint(view_comment)
        foreign_key_constraints = reflect_virtual_foreign_key_constraints(view_comment)

        # Reflect view as base table
        view_tab = Table(
            view_name,
            declarative_base.metadata,
            schema=schema,
            autoload=True,
            autoload_with=engine,
            *[primary_key_constraint],
            *foreign_key_constraints,
        )

        class_name = rename_table(declarative_base, view_name, view_tab)

        # ORM View Table
        view_orm = type(
            class_name,
            (
                declarative_base,
                ViewMixin,
            ),
            {"__table__": view_tab},
        )
        views.append(view_orm)  # type: ignore

    return views


def reflect_virtual_primary_key_constraint(comment: str) -> PrimaryKeyConstraint:
    """
    Reflects virtual primary key constraints from view comments

    Format:
        @primary_key (id, key2)'
    """
    for row in comment.split("\n"):
        if not row.startswith("@primary_key"):
            continue

        # Remove all spaces
        # Ex: @primary_key(id)
        row = row.replace(" ", "")

        # Ex: (variant_id)referencespublic.variant(id)
        col_names = row.split("(")[1].strip().strip(")").split(",")
        return PrimaryKeyConstraint(*col_names)

    raise Exception("Views must have a @primary_key comment")


def reflect_virtual_foreign_key_constraints(comment: str) -> List[ForeignKeyConstraint]:
    """
    Reflects virtual foreign key constraints from view comments

    Format:
        @foreign_key (variant_id, key2) references public.variant (id, key2)'

    """
    foreign_keys = []
    for row in comment.split("\n"):
        if not row.startswith("@foreign_key"):
            continue

        # Remove all spaces
        # Ex: @foreign_key(variant_id)referencespublic.variant(id)
        row = row.replace(" ", "")

        # Ex: (variant_id)referencespublic.variant(id)
        row = row.strip("@foreign_key")

        # Ex: ('(variant_id', 'public.variant(id)')
        local, remote = row.split(")references")

        # Ex: ['variant_id']
        local_col_names = local.lstrip("(").split(",")

        # Ex: ('public.variant', 'id)')
        remote_table, remote = remote.lstrip("references").split("(")
        remote_col_names = remote.strip().rstrip(")").split(",")

        remote_qualified_col_names = [f"{remote_table}.{remote_col_name}" for remote_col_name in remote_col_names]
        foreign_key = ForeignKeyConstraint(local_col_names, remote_qualified_col_names)
        foreign_keys.append(foreign_key)
    return foreign_keys



from functools import lru_cache

from starlette.requests import Request
from starlette.responses import HTMLResponse


async def graphiql_endpoint(request: Request, graphql_url_path: str = '"/"') -> HTMLResponse:
    """Return the HTMLResponse for GraphiQL GraphQL explorer configured to hit the correct endpoint"""

    html_text = build_graphiql_html(graphql_url_path)
    return HTMLResponse(html_text)


@lru_cache()
def build_graphiql_html(url_path: str) -> str:
    """Return the raw HTML for GraphiQL GraphQL explorer"""

    text = GRAPHIQL.replace("{{REQUEST_PATH}}", url_path)
    return text


GRAPHIQL = """
<html>
  <head>
    <title>Nebulo GraphiQL</title>
    <link href="static/graphiql.min.css" rel="stylesheet" />
  </head>
  <body style="margin: 0;">
    <div id="graphiql" style="height: 100vh;"></div>

    <script
      crossorigin
      src="https://unpkg.com/react/umd/react.production.min.js"
    ></script>
    <script
      crossorigin
      src="https://unpkg.com/react-dom/umd/react-dom.production.min.js"
    ></script>
    <script
      crossorigin
      src="static/graphiql.min.js"
    ></script>

    <script>

        let graphQLFetcher = (graphQLParams, opts = { headers: {} }) => {
          return fetch(window.location.origin + {{REQUEST_PATH}}, {
            method: 'post',
            headers: Object.assign({ 'Content-Type': 'application/json', 'User-Agent': 'GraphiQL' }, opts.headers),
            body: JSON.stringify(graphQLParams),
          }).then(response => response.json());
        }

        ReactDOM.render(
          React.createElement(GraphiQL, {
            fetcher: graphQLFetcher,
            headerEditorEnabled: true,
          }),
        document.getElementById('graphiql'),
      );
    </script>
  </body>
</html>
"""



# pylint: disable=invalid-name
# mypy: ignore-errors
from __future__ import annotations

import typing
from functools import lru_cache

from flupy import flu
from nebulo.config import Config
from nebulo.gql.alias import CompositeType, ConnectionType, ScalarType, TableType
from nebulo.gql.parse_info import ASTNode
from nebulo.gql.relay.cursor import to_cursor_sql
from nebulo.gql.relay.node_interface import ID, to_node_id_sql
from nebulo.sql.inspect import get_columns, get_primary_key_columns, get_relationships, get_table_name
from nebulo.sql.sanitize import secure_random_string
from nebulo.sql.table_base import TableProtocol
from sqlalchemy import Column, Integer, and_, asc, cast, desc, func, literal, literal_column, select, tuple_
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import RelationshipProperty
from sqlalchemy.sql import Alias, Select
from sqlalchemy.sql.elements import BinaryExpression, Label


def sql_builder(tree: ASTNode, parent_name: typing.Optional[str] = None) -> Alias:
    return_type = tree.return_type

    if isinstance(return_type, TableType):
        return row_block(field=tree, parent_name=parent_name)

    if isinstance(return_type, ConnectionType):
        return connection_block(field=tree, parent_name=parent_name)

    # SQL Function handler for immutable functions
    if hasattr(return_type, "sql_function"):
        # Immutable function
        sql_func_callable = return_type.sql_function.to_executable(tree.args.values())
        return select([sql_func_callable.label("ret_json")]).alias()

    raise Exception("sql builder could not match return type")


@lru_cache()
def field_name_to_column(sqla_model: TableProtocol, gql_field_name: str) -> Column:
    for column in get_columns(sqla_model):
        if Config.column_name_mapper(column) == gql_field_name:
            return column
    raise KeyError(f"No column corresponding to field {gql_field_name}")


@lru_cache()
def field_name_to_relationship(sqla_model: TableProtocol, gql_field_name: str) -> RelationshipProperty:
    for relationship in get_relationships(sqla_model):
        if Config.relationship_name_mapper(relationship) == gql_field_name:
            return relationship
    raise Exception(f"No relationship corresponding to field {gql_field_name}")


def to_join_clause(field: ASTNode, parent_block_name: str) -> typing.List[BinaryExpression]:
    parent_field = field.parent
    assert parent_field is not None
    relation_from_parent = field_name_to_relationship(parent_field.return_type.sqla_model, field.name)
    local_table_name = get_table_name(field.return_type.sqla_model)

    join_clause: typing.List[BinaryExpression] = []
    for parent_col, local_col in relation_from_parent.local_remote_pairs:
        parent_col_name = parent_col.name
        local_col_name = local_col.name
        join_clause.append(
            literal_column(f"{parent_block_name}.{parent_col_name}")
            == literal_column(f"{local_table_name}.{local_col_name}")
        )
    return join_clause


def to_pkey_clause(field: ASTNode, pkey_eq: typing.List[str]) -> typing.List[BinaryExpression]:
    local_table = field.return_type.sqla_model
    local_table_name = get_table_name(field.return_type.sqla_model)
    pkey_cols = get_primary_key_columns(local_table)

    res = []
    for col, val in zip(pkey_cols, pkey_eq):
        res.append(literal_column(f"{local_table_name}.{col.name}") == val)
    return res


def to_limit(field: ASTNode) -> int:
    args = field.args
    default = 20
    first = int(args.get("first", default))
    last = int(args.get("last", default))
    limit = min(first, last, default)
    return limit


def to_conditions_clause(field: ASTNode) -> typing.List[BinaryExpression]:
    return_sqla_model = field.return_type.sqla_model
    local_table_name = get_table_name(return_sqla_model)
    args = field.args

    conditions = args.get("condition")

    if conditions is None:
        return [True]

    res = []
    for field_name, val in conditions.items():
        column_name = field_name_to_column(return_sqla_model, field_name).name
        res.append(literal_column(f"{local_table_name}.{column_name}") == val)
    return res


def build_relationship(field: ASTNode, block_name: str) -> Label:
    return sql_builder(field, block_name).as_scalar().label(field.alias)


def literal_string(text):
    return literal_column(f"'{text}'")


def sql_finalize(return_name: str, expr: Alias) -> Select:
    final = select([func.jsonb_build_object(literal_string(return_name), expr.c.ret_json).label("json")]).select_from(
        expr
    )
    return final


def row_block(field: ASTNode, parent_name: typing.Optional[str] = None) -> Alias:
    return_type = field.return_type
    sqla_model = return_type.sqla_model
    core_model = sqla_model.__table__

    block_name = secure_random_string()
    if parent_name is None:
        # If there is no parent, nodeId is mandatory
        pkey_cols = get_primary_key_columns(sqla_model)
        node_id = field.args["nodeId"]
        pkey_clause = [col == node_id.values[str(col.name)] for col in pkey_cols]
        join_clause = [True]
    else:
        # If there is a parent no arguments are accepted
        join_clause = to_join_clause(field, parent_name)
        pkey_clause = [True]

    core_model_ref = (select(core_model.c).where(and_(*pkey_clause, *join_clause))).alias(block_name)

    select_clause = []
    for subfield in field.fields:

        if subfield.return_type == ID:
            elem = select([to_node_id_sql(sqla_model, core_model_ref)]).label(subfield.alias)
            select_clause.append(elem)
        elif isinstance(subfield.return_type, (ScalarType, CompositeType)):
            col_name = field_name_to_column(sqla_model, subfield.name).name
            elem = core_model_ref.c[col_name].label(subfield.alias)
            select_clause.append(elem)
        else:
            elem = build_relationship(subfield, block_name)
            select_clause.append(elem)

    block = (
        select(
            [
                func.jsonb_build_object(
                    *flu(select_clause).map(lambda x: (literal_string(x.key), x)).flatten().collect()
                ).label("ret_json")
            ]
        ).select_from(core_model_ref)
    ).alias()

    return block


def check_has_total(field: ASTNode) -> bool:
    "Check if 'totalCount' is requested in the query result set"
    return any(x.name in "totalCount" for x in field.fields)


def get_edge_node_fields(field):
    """Returns connection.edge.node fields"""
    for cfield in field.fields:
        if cfield.name == "edges":
            for edge_field in cfield.fields:
                if edge_field.name == "node":
                    return edge_field.fields
    return []


ONE = literal_column("1")
ZERO = literal_column("0")
TRUE = literal_column("true")
FALSE = literal_column("false")


def connection_block(field: ASTNode, parent_name: typing.Optional[str]) -> Alias:
    return_type = field.return_type
    sqla_model = return_type.sqla_model

    block_name = secure_random_string()
    if parent_name is None:
        join_conditions = [True]
    else:
        join_conditions = to_join_clause(field, parent_name)

    filter_conditions = to_conditions_clause(field)
    limit = to_limit(field)
    has_total = check_has_total(field)

    is_page_after = "after" in field.args
    is_page_before = "before" in field.args

    totalCount_alias = field.get_subfield_alias(["totalCount"])

    edges_alias = field.get_subfield_alias(["edges"])
    node_alias = field.get_subfield_alias(["edges", "node"])
    cursor_alias = field.get_subfield_alias(["edges", "cursor"])

    pageInfo_alias = field.get_subfield_alias(["pageInfo"])
    hasNextPage_alias = field.get_subfield_alias(["pageInfo", "hasNextPage"])
    hasPreviousPage_alias = field.get_subfield_alias(["pageInfo", "hasPreviousPage"])
    startCursor_alias = field.get_subfield_alias(["pageInfo", "startCursor"])
    endCursor_alias = field.get_subfield_alias(["pageInfo", "endCursor"])

    # Apply Filters
    core_model = sqla_model.__table__
    core_model_ref = (
        select(core_model.c)
        .select_from(core_model)
        .where(
            and_(
                # Join clause
                *join_conditions,
                # Conditions
                *filter_conditions,
            )
        )
    ).alias(block_name)

    new_edge_node_selects = []
    new_relation_selects = []

    for subfield in get_edge_node_fields(field):
        # Does anything other than NodeID go here?
        if subfield.return_type == ID:
            # elem = select([to_node_id_sql(sqla_model, core_model_ref)]).label(subfield.alias)
            elem = to_node_id_sql(sqla_model, core_model_ref).label(subfield.alias)
            new_edge_node_selects.append(elem)
        elif isinstance(subfield.return_type, (ScalarType, CompositeType)):
            col_name = field_name_to_column(sqla_model, subfield.name).name
            elem = core_model_ref.c[col_name].label(subfield.alias)
            new_edge_node_selects.append(elem)
        else:
            elem = build_relationship(subfield, block_name)
            new_relation_selects.append(elem)

    # Setup Pagination
    args = field.args
    after_cursor = args.get("after", None)
    before_cursor = args.get("before", None)
    first = args.get("first", None)
    last = args.get("last", None)

    if first is not None and last is not None:
        raise ValueError('only one of "first" and "last" may be provided')

    if after_cursor or before_cursor:
        local_table_name = get_table_name(field.return_type.sqla_model)
        cursor_table_name = before_cursor.table_name if before_cursor else after_cursor.table_name
        cursor_values = before_cursor.values if before_cursor else after_cursor.values

        if after_cursor is not None and before_cursor is not None:
            raise ValueError('only one of "before" and "after" may be provided')

        if after_cursor is not None and last is not None:
            raise ValueError('"after" is not compatible with "last". Use "first"')

        if before_cursor is not None and first is not None:
            raise ValueError('"before" is not compatible with "first". Use "last"')

        if cursor_table_name != local_table_name:
            raise ValueError("Invalid cursor for entity type")

        pkey_cols = get_primary_key_columns(sqla_model)

        pagination_clause = tuple_(*[core_model_ref.c[col.name] for col in pkey_cols]).op(
            ">" if after_cursor is not None else "<"
        )(tuple_(*[cursor_values[col.name] for col in pkey_cols]))
    else:
        pagination_clause = True

    order_clause = [asc(core_model_ref.c[col.name]) for col in get_primary_key_columns(sqla_model)]
    reverse_order_clause = [desc(core_model_ref.c[col.name]) for col in get_primary_key_columns(sqla_model)]

    total_block = (
        select([func.count(ONE).label("total_count")]).select_from(core_model_ref.alias()).where(has_total)
    ).alias(block_name + "_total")

    node_id_sql = to_node_id_sql(sqla_model, core_model_ref)
    cursor_sql = to_cursor_sql(sqla_model, core_model_ref)

    # Select the right stuff
    p1_block = (
        select(
            [
                *new_edge_node_selects,
                *new_relation_selects,
                # For internal Use
                node_id_sql.label("_nodeId"),
                cursor_sql.label("_cursor"),
                # For internal Use
                func.row_number().over().label("_row_num"),
            ]
        )
        .select_from(core_model_ref)
        .where(pagination_clause)
        .order_by(*(reverse_order_clause if is_page_before else order_clause), *order_clause)
        .limit(cast(limit + 1, Integer()))
    ).alias(block_name + "_p1")

    # Drop maybe extra row
    p2_block = (
        select([*p1_block.c, (func.max(p1_block.c._row_num).over() > limit).label("_has_next_page")])
        .select_from(p1_block)
        .limit(limit)
    ).alias(block_name + "_p2")

    ordering = desc(literal_column("_row_num")) if is_page_before else asc(literal_column("_row_num"))

    p3_block = (select(p2_block.c).select_from(p2_block).order_by(ordering)).alias(block_name + "_p3")

    final = (
        select(
            [
                func.jsonb_build_object(
                    literal_string(totalCount_alias),
                    func.coalesce(func.min(total_block.c.total_count), ZERO) if has_total else None,
                    literal_string(pageInfo_alias),
                    func.jsonb_build_object(
                        literal_string(hasNextPage_alias),
                        func.coalesce(func.array_agg(p3_block.c._has_next_page)[ONE], FALSE),
                        literal_string(hasPreviousPage_alias),
                        TRUE if is_page_after else FALSE,
                        literal_string(startCursor_alias),
                        func.array_agg(p3_block.c._nodeId)[ONE],
                        literal_string(endCursor_alias),
                        func.array_agg(p3_block.c._nodeId)[func.array_upper(func.array_agg(p3_block.c._nodeId), ONE)],
                    ),
                    literal_string(edges_alias),
                    func.coalesce(
                        func.jsonb_agg(
                            func.jsonb_build_object(
                                literal_string(cursor_alias),
                                p3_block.c._nodeId,
                                literal_string(node_alias),
                                func.cast(func.row_to_json(literal_column(p3_block.name)), JSONB()),
                            )
                        ),
                        func.cast(literal("[]"), JSONB()),
                    ),
                ).label("ret_json")
            ]
        )
        .select_from(p3_block)
        .select_from(total_block if has_total else select([1]).alias())
    ).alias()

    return final



# pylint: disable=unused-argument,invalid-name,line-too-long,unsubscriptable-object
from __future__ import annotations

from itertools import zip_longest
from typing import Any, List, Optional, Type

from nebulo.exceptions import SQLParseError
from sqlalchemy import cast, func
from sqlalchemy import text as sql_text
from sqlalchemy.sql import sqltypes
from sqlalchemy.sql.type_api import TypeEngine


class SQLFunction:
    """A PostgreSQL Function

    INTERNAL USE ONLY
    """

    def __init__(
        self,
        schema: str,
        name: str,
        arg_names: List[Optional[str]],
        arg_pg_types: List[str],
        arg_sqla_types: List[Type[TypeEngine[Any]]],
        return_sqla_type: Type[TypeEngine[Any]],
        return_pg_type_schema: str,
        return_pg_type: str,
        is_immutable: bool,
    ):
        if len(arg_names) != len(arg_sqla_types) != len(arg_pg_types):
            raise SQLParseError("SQLFunction requires same number of arg_names and sqla_types")
        self.schema = schema
        self.name = name
        self.arg_names = arg_names
        self.arg_pg_types = arg_pg_types
        self.arg_sqla_types = arg_sqla_types
        self.return_sqla_type = return_sqla_type
        self.return_pg_type_schema = return_pg_type_schema
        self.return_pg_type = return_pg_type
        self.is_immutable = is_immutable

    def to_executable(self, args: List):

        if len(args) != len(self.arg_pg_types):
            raise SQLParseError(f"Invalid number of parameters for SQLFunction {self.schema}.{self.name}")

        # SQLAlchemy dynamic representation of the function
        sqla_func = getattr(getattr(func, self.schema), self.name)
        # Bind and typecast user parameters
        call_sig = [cast(arg_value, arg_sqla_type) for arg_value, arg_sqla_type in zip(args, self.arg_sqla_types)]
        return sqla_func(*call_sig)


def reflect_functions(engine, schema, type_map) -> List[SQLFunction]:
    """Get a list of functions available in the database"""

    # TODO: Support default arguments
    # I haven't been able to find a way to get an array of default args
    # but you can get a function signature including and not including
    # see proargdefaults, pronargdefaults,
    # pg_get_function_identity_arguments, get_function_arguments

    sql = sql_text(
        """
    with extension_functions as (
        select
            objid as extension_function_oid
        from
            pg_depend
        where
            -- depends on an extension
            deptype='e'
            -- is a proc/function
            and classid = 'pg_proc'::regclass
    )

    select
        n.nspname as function_schema,
        p.proname as function_name,
        proargnames arg_names,
        (select array_agg((select typnamespace::regnamespace::text from pg_type where oid=type_oid)) from unnest(proargtypes) x(type_oid)) arg_types_schema,
        (select array_agg(type_oid::regtype::text) from unnest(proargtypes) x(type_oid)) arg_types,
        t.typnamespace::regnamespace::text as return_type_schema,
        t.typname as return_type,
        p.provolatile = 'i' is_immutable
    from
        pg_proc p
        left join pg_namespace n on p.pronamespace = n.oid
        left join pg_language l on p.prolang = l.oid
        left join pg_type t on t.oid = p.prorettype
        left join extension_functions ef on p.oid = ef.extension_function_oid
    where
        n.nspname not in ('pg_catalog', 'information_schema')
        and ef.extension_function_oid is null
        and n.nspname like :schema
        """
    )
    rows = engine.execute(sql, schema=schema).fetchall()

    functions: List[SQLFunction] = []

    for (
        func_schema,
        func_name,
        arg_names,
        arg_type_schemas,
        pg_arg_types,
        pg_return_type_schema,
        pg_return_type_name,
        is_immutable,
    ) in rows:
        arg_names = arg_names or []
        pg_arg_types = pg_arg_types or []
        pg_arg_names = [arg_name for arg_name, _ in zip_longest(arg_names, pg_arg_types, fillvalue=None)]
        arg_type_schemas = arg_type_schemas or []
        sqla_arg_types = [
            type_map.get(pg_type_name, sqltypes.NULLTYPE)
            for pg_type_schema, pg_type_name in zip(arg_type_schemas, pg_arg_types) or []
        ]
        sqla_return_type = type_map.get(pg_return_type_name, sqltypes.NULLTYPE)

        function = SQLFunction(
            schema=func_schema,
            name=func_name,
            arg_names=pg_arg_names,
            arg_pg_types=pg_arg_types,
            arg_sqla_types=sqla_arg_types,
            return_sqla_type=sqla_return_type,
            return_pg_type_schema=pg_return_type_schema,
            return_pg_type=pg_return_type_name,
            is_immutable=is_immutable,
        )
        functions.append(function)

    return functions



import os
from pathlib import Path
from typing import Callable, List, Optional

from databases import Database
from nebulo.gql.alias import Schema
from nebulo.gql.sqla_to_gql import sqla_models_to_graphql_schema
from nebulo.server.exception import http_exception
from nebulo.server.routes import get_graphql_endpoint, graphiql_endpoint
from nebulo.sql.reflection.manager import reflect_sqla_models
from sqlalchemy import Table, create_engine
from starlette.applications import Starlette
from starlette.exceptions import HTTPException
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
from starlette.routing import Mount, Route
from starlette.staticfiles import StaticFiles

STATIC_PATH = Path(os.path.abspath(__file__)).parent.parent.resolve() / "static"


def create_app(
    connection: str,
    schema: str = "public",
    jwt_identifier: Optional[str] = None,
    jwt_secret: Optional[str] = None,
    before_start: Optional[Callable[[Schema, List[Table], Database], Schema]] = None,
) -> Starlette:
    """Create an ASGI App"""

    if not (jwt_identifier is not None) == (jwt_secret is not None):
        raise Exception("jwt_token_identifier and jwt_secret must be provided together")

    database = Database(connection)
    # Reflect database to sqla models
    sqla_engine = create_engine(connection)
    sqla_models, sql_functions = reflect_sqla_models(engine=sqla_engine, schema=schema)

    # Convert sqla models to graphql schema
    gql_schema = sqla_models_to_graphql_schema(
        sqla_models,
        sql_functions,
        jwt_identifier=jwt_identifier,
        jwt_secret=jwt_secret,
    )

    if before_start:
        core_tables = [x.__table__ for x in sqla_models]
        gql_schema = before_start(gql_schema, core_tables, database)

    # Build Starlette app
    graphql_endpoint = get_graphql_endpoint(gql_schema, database, jwt_secret)

    routes = [
        Route("/", graphql_endpoint, methods=["POST"]),
        Route("/graphiql", graphiql_endpoint, methods=["GET"]),
        Mount("/static", StaticFiles(directory=STATIC_PATH), name="static"),
    ]

    middleware = [Middleware(CORSMiddleware, allow_origins=["*"])]

    _app = Starlette(
        routes=routes,
        middleware=middleware,
        exception_handlers={HTTPException: http_exception},
        on_startup=[database.connect],
        on_shutdown=[database.disconnect],
    )

    return _app



from __future__ import annotations

from typing import List, Tuple

from nebulo.sql.reflection.function import SQLFunction, reflect_functions
from nebulo.sql.reflection.names import rename_table, rename_to_many_collection, rename_to_one_collection
from nebulo.sql.reflection.types import reflect_composites
from nebulo.sql.reflection.views import reflect_views
from nebulo.sql.table_base import TableProtocol
from sqlalchemy import MetaData, types
from sqlalchemy.dialects.postgresql import base as pg_base
from sqlalchemy.engine import Engine
from sqlalchemy.ext.automap import automap_base


def reflect_sqla_models(engine: Engine, schema: str = "public") -> Tuple[List[TableProtocol], List[SQLFunction]]:
    """Reflect SQLAlchemy Declarative Models from a database connection"""

    meta = MetaData()
    declarative_base = automap_base(metadata=meta)

    # Register event listeners to apply GQL attr keys to columns

    # Retrive a copy of the full type map
    basic_type_map = pg_base.ischema_names.copy()

    # Reflect composite types (not supported by sqla)
    composites = reflect_composites(engine, schema, basic_type_map)

    # Register composite types with SQLA to make them available during reflection
    pg_base.ischema_names.update({type_name: type_ for (type_schema, type_name), type_ in composites.items()})

    # Retrive a copy of the full type map
    # NOTE: types are not schema namespaced so colisions can occur reflecting tables
    type_map = pg_base.ischema_names.copy()
    type_map["bool"] = types.Boolean  # type: ignore

    # Reflect views as SQLA ORM table
    views = reflect_views(engine=engine, schema=schema, declarative_base=declarative_base)

    declarative_base.prepare(
        engine,
        reflect=True,
        schema=schema,
        classname_for_table=rename_table,
        name_for_scalar_relationship=rename_to_one_collection,
        name_for_collection_relationship=rename_to_many_collection,
    )

    # Register tables as types so functions can return a row of a table
    tables = list(declarative_base.classes) + views
    for table in tables + views:
        type_map[table.__table__.name] = table

    # Reflect functions, allowing composite types
    functions = reflect_functions(engine=engine, schema=schema, type_map=type_map)

    # SQLA Tables
    return (tables, functions)



from __future__ import annotations

from functools import lru_cache

from nebulo.config import Config
from nebulo.gql.alias import (
    DeleteInputType,
    DeletePayloadType,
    Field,
    InputObjectType,
    NonNull,
    ObjectType,
    String,
    TableInputType,
)
from nebulo.gql.relay.node_interface import ID
from nebulo.gql.resolve.resolvers.default import default_resolver
from nebulo.sql.table_base import TableProtocol

"""
deleteAccount(input: DeleteAccountInput!):
    deleteAccountPayload
"""


@lru_cache()
def delete_entrypoint_factory(sqla_model: TableProtocol, resolver) -> Field:
    """deleteAccount"""
    relevant_type_name = Config.table_type_name_mapper(sqla_model)
    name = f"delete{relevant_type_name}"
    args = {"input": NonNull(delete_input_type_factory(sqla_model))}
    payload = delete_payload_factory(sqla_model)
    return {
        name: Field(
            payload,
            args=args,
            resolve=resolver,
            description=f"Delete a single {relevant_type_name} using its globally unique id and a patch.",
        )
    }


@lru_cache()
def delete_input_type_factory(sqla_model: TableProtocol) -> InputObjectType:
    """DeleteAccountInput!"""
    relevant_type_name = Config.table_type_name_mapper(sqla_model)
    result_name = f"Delete{relevant_type_name}Input"

    input_object_name = Config.table_name_mapper(sqla_model)

    attrs = {
        "nodeId": NonNull(ID),
        "clientMutationId": String,
    }
    return DeleteInputType(result_name, attrs, description=f"All input for the create {relevant_type_name} mutation.")


@lru_cache()
def delete_payload_factory(sqla_model: TableProtocol) -> InputObjectType:
    """DeleteAccountPayload"""

    relevant_type_name = Config.table_type_name_mapper(sqla_model)
    result_name = f"Delete{relevant_type_name}Payload"

    attrs = {"clientMutationId": Field(String, resolve=default_resolver), "nodeId": ID}

    return DeletePayloadType(
        result_name, attrs, description=f"The output of our delete {relevant_type_name} mutation", sqla_model=sqla_model
    )



"""
When SQL functions are marked as immutable, they are handled as queries
Otherwise, they are mutations
"""
from __future__ import annotations

import typing
from functools import lru_cache

import jwt
from nebulo.config import Config
from nebulo.gql.alias import (
    Argument,
    Field,
    FunctionInputType,
    FunctionPayloadType,
    InputObjectType,
    NonNull,
    ObjectType,
    ScalarType,
    String,
    TableInputType,
)
from nebulo.gql.convert.column import convert_input_type, convert_type
from nebulo.gql.resolve.resolvers.default import default_resolver
from nebulo.sql.reflection.function import SQLFunction

__all__ = ["mutable_function_entrypoint_factory", "immutable_function_entrypoint_factory"]

"""
# When volatile
authenticate(input: AuthenticateInput!):
    AuthenticatePayload

# When immutable
toUpper(someText: String!):
    String
"""


@lru_cache()
def immutable_function_entrypoint_factory(
    sql_function: SQLFunction, resolver: typing.Callable
) -> typing.Dict[str, Field]:
    """authenticate"""
    # TODO(OR): need seperate mapper
    function_name = Config.function_name_mapper(sql_function)

    if not sql_function.is_immutable:
        raise Exception(f"SQLFunction {sql_function.name} is not immutable, use mutable_function_entrypoint")

    gql_args = {
        (arg_name if arg_name else f"param{ix}"): Argument(NonNull(convert_type(arg_sqla_type)))
        for ix, (arg_name, arg_sqla_type) in enumerate(zip(sql_function.arg_names, sql_function.arg_sqla_types))
    }

    return_type = convert_type(sql_function.return_sqla_type)
    return_type.sql_function = sql_function
    return_field = Field(return_type, args=gql_args, resolve=resolver, description="")

    return {function_name: return_field}


def is_jwt_function(sql_function: SQLFunction, jwt_identifier: typing.Optional[str]):
    function_return_type_identifier = sql_function.return_pg_type_schema + "." + sql_function.return_pg_type
    return function_return_type_identifier == jwt_identifier


@lru_cache()
def mutable_function_entrypoint_factory(
    sql_function: SQLFunction, resolver: typing.Callable, jwt_secret: typing.Optional[str] = None
) -> typing.Dict[str, Field]:
    """authenticate"""
    # TODO(OR): need seperate mapper
    function_name = Config.function_name_mapper(sql_function)

    if sql_function.is_immutable:
        raise Exception(f"SQLFunction {sql_function.name} is immutable, use immutable_function_entrypoint")

    args = {"input": NonNull(function_input_type_factory(sql_function))}

    if jwt_secret is not None:
        payload = jwt_function_payload_factory(sql_function, jwt_secret)
    else:
        payload = function_payload_factory(sql_function)

    return {
        function_name: Field(payload, args=args, resolve=resolver, description=f"Call the function {function_name}.")
    }


# Remainder relates to mutation functions


@lru_cache()
def function_input_type_factory(sql_function: SQLFunction) -> FunctionInputType:
    """AuthenticateInput!"""
    function_name = Config.function_type_name_mapper(sql_function)
    result_name = f"{function_name}Input"

    function_args = {
        (arg_name if arg_name else f"param{ix}"): NonNull(convert_input_type(arg_sqla_type))
        for ix, (arg_name, arg_sqla_type) in enumerate(zip(sql_function.arg_names, sql_function.arg_sqla_types))
    }

    attrs = {"clientMutationId": String, **function_args}
    return FunctionInputType(result_name, attrs, description=f"All input for the {function_name} mutation.")


@lru_cache()
def function_payload_factory(sql_function: SQLFunction) -> FunctionPayloadType:
    """CreateAccountPayload"""
    function_name = Config.function_type_name_mapper(sql_function)
    result_name = f"{function_name}Payload"

    # TODO(OR): handle functions with no return
    function_return_type = convert_type(sql_function.return_sqla_type)
    function_return_type.sql_function = sql_function

    attrs = {
        "clientMutationId": Field(String, resolve=default_resolver),
        "result": Field(
            function_return_type,
            description=f"The {result_name} that was created by this mutation.",
            resolve=default_resolver,
        ),
    }

    payload = FunctionPayloadType(result_name, attrs, description=f"The output of our create {function_name} mutation")
    payload.sql_function = sql_function
    return payload


@lru_cache()
def jwt_function_payload_factory(sql_function: SQLFunction, jwt_secret: str) -> FunctionPayloadType:
    """CreateAccountPayload"""
    function_name = Config.function_type_name_mapper(sql_function)
    result_name = f"{function_name}Payload"

    function_return_type = ScalarType(
        "JWT",
        serialize=lambda result: jwt.encode({k: v for k, v in result.items()}, jwt_secret, algorithm="HS256").decode(
            "utf-8"
        ),
    )

    attrs = {
        "clientMutationId": Field(String, resolve=default_resolver),
        "result": Field(
            function_return_type,
            description=f"The {result_name} that was created by this mutation.",
            resolve=default_resolver,
        ),
    }

    payload = FunctionPayloadType(result_name, attrs, description=f"The output of our create {function_name} mutation")
    payload.sql_function = sql_function
    return payload



from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import JSONResponse


async def http_exception(request: Request, exc: HTTPException):
    return JSONResponse({"data": None, "errors": [exc.detail]}, status_code=exc.status_code)



from typing import Any, Awaitable, Callable, Dict, Optional

import jwt
from jwt.exceptions import DecodeError, ExpiredSignatureError, PyJWTError
from starlette.exceptions import HTTPException
from starlette.requests import Request


def get_jwt_claims_handler(secret: Optional[str]) -> Callable[[Request], Awaitable[Dict[str, Any]]]:
    """Return a function that retrieves and decodes JWT claims from the Starlette Request"""

    async def get_jwt_claims(request: Request) -> Dict[str, Any]:
        """Retrieve the JWT claims from the Starlette Request"""

        if secret is None:
            return {}

        if "Authorization" not in request.headers:
            return {}

        auth = request.headers["Authorization"]
        try:
            scheme, token = auth.split()
            if scheme.lower() == "bearer":
                contents = jwt.decode(token, secret, algorithms=["HS256"])
                return contents
        except ValueError:
            # The user probably forgot to prepend "Bearer "
            raise HTTPException(401, "Invalid JWT Authorization header. Expected 'Bearer <JWT>'")
        except DecodeError:
            raise HTTPException(401, "Invalid JWT credentials")
        except ExpiredSignatureError:
            raise HTTPException(401, "JWT has expired. Please reauthenticate")
        # Generically catch all PyJWT errors
        except PyJWTError as exc:
            raise HTTPException(401, str(exc))
        return {}

    return get_jwt_claims



# pylint: disable=invalid-name,unsubscriptable-object
from __future__ import annotations

import typing
from functools import lru_cache

from nebulo.gql.alias import (
    Boolean,
    CompositeType,
    Field,
    Float,
    InputField,
    InputObjectType,
    Int,
    NonNull,
    ScalarType,
    String,
)
from nebulo.gql.resolve.resolvers.default import default_resolver
from nebulo.sql.composite import CompositeType as SQLACompositeType
from nebulo.sql.table_base import TableProtocol
from nebulo.text_utils import snake_to_camel
from sqlalchemy import Column, types
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql.type_api import TypeEngine

UnknownType = ScalarType(name="UnknownString", serialize=str)
DateTimeType = ScalarType(name="DateTime", serialize=str)
DateType = ScalarType(name="Date", serialize=str)
TimeType = ScalarType(name="Time", serialize=str)
UUIDType = ScalarType(name="UUID", serialize=str)
INETType = ScalarType(name="INET", serialize=str)
CIDRType = ScalarType(name="CIDR", serialize=str)


SQLA_TO_GQL = {
    types.Boolean: Boolean,
    # Integer
    types.Integer: Int,
    types.INTEGER: Int,
    types.BIGINT: Int,
    # Float
    types.Float: Float,
    types.Numeric: Float,
    # required b/c json uses 32 bit integers
    types.BigInteger: Float,
    # Text
    types.Text: String,
    types.String: String,
    types.Unicode: String,
    types.UnicodeText: String,
    # Date
    types.Date: DateType,
    types.Time: TimeType,
    types.DateTime: DateTimeType,
    postgresql.TIMESTAMP: DateTimeType,
    # Other
    postgresql.UUID: UUIDType,
    postgresql.INET: INETType,
    postgresql.CIDR: CIDRType,
}


@lru_cache()
def convert_type(sqla_type: typing.Type[TypeEngine]):
    if issubclass(sqla_type, SQLACompositeType):
        return composite_factory(sqla_type)

    if isinstance(sqla_type, TableProtocol):
        from .table import table_factory

        return table_factory(sqla_type)

    # TODO(OR): Enums
    return SQLA_TO_GQL.get(sqla_type, String)


@lru_cache()
def convert_column(column: Column) -> Field:
    """Converts a sqlalchemy column into a graphql field or input field"""
    sqla_type = type(column.type)
    gql_type = convert_type(sqla_type)
    notnull = not column.nullable
    return_type = NonNull(gql_type) if notnull else gql_type
    return Field(return_type, resolve=default_resolver)


@lru_cache()
def composite_factory(sqla_composite: SQLACompositeType) -> CompositeType:
    name = snake_to_camel(sqla_composite.name, upper=True)
    fields = {}

    for column in sqla_composite.columns:
        column_key = str(column.key)
        gql_key = column_key
        gql_type = convert_column(column)
        fields[gql_key] = gql_type

    return_type = CompositeType(name, fields)
    return_type.sqla_composite = sqla_composite
    return return_type


@lru_cache()
def convert_input_type(sqla_type: typing.Type[TypeEngine[typing.Any]]):
    if issubclass(sqla_type, SQLACompositeType):
        gql_type = composite_input_factory(sqla_type)
    else:
        gql_type = SQLA_TO_GQL.get(sqla_type, String)
    return gql_type


@lru_cache()
def convert_column_to_input(column: Column) -> InputField:
    """Converts a sqlalchemy column into a graphql field or input field"""
    sqla_type = type(column.type)
    gql_type = convert_input_type(sqla_type)
    return InputField(gql_type)


@lru_cache()
def composite_input_factory(sqla_composite: SQLACompositeType) -> InputObjectType:
    name = snake_to_camel(sqla_composite.name, upper=True) + "Input"
    fields = {}

    for column in sqla_composite.columns:
        column_key = str(column.key)
        gql_key = column_key
        gql_field = convert_column_to_input(column)
        fields[gql_key] = gql_field

    return_type = InputObjectType(name, fields)
    return_type.sqla_composite = sqla_composite
    return return_type



from functools import lru_cache

from inflect import engine


@lru_cache()
def get_pluralizer() -> engine:
    """Return an instance of inflection library's engine.
    This is wrapped in a function to reduce import side effects"""
    return engine()


@lru_cache()
def to_plural(text: str) -> str:
    pluralizer = get_pluralizer()
    return pluralizer.plural(text)



# pylint: disable=invalid-name
from typing import Dict, Tuple

from flupy import flu
from nebulo.sql.composite import composite_type_factory
from nebulo.text_utils import snake_to_camel
from sqlalchemy import Column
from sqlalchemy import text as sql_text
from sqlalchemy.sql import sqltypes
from sqlalchemy.sql.type_api import TypeEngine


def reflect_composites(engine, schema, type_map) -> Dict[Tuple[str, str], TypeEngine]:
    """Get a list of functions available in the database"""

    sql = sql_text(
        """
WITH types AS (
    SELECT
        n.nspname,
        pg_catalog.format_type ( t.oid, NULL ) AS obj_name,
        CASE
            WHEN t.typrelid != 0 THEN CAST ( 'tuple' AS pg_catalog.text )
            WHEN t.typlen < 0 THEN CAST ( 'var' AS pg_catalog.text )
            ELSE CAST ( t.typlen AS pg_catalog.text )
            END AS obj_type,
        coalesce ( pg_catalog.obj_description ( t.oid, 'pg_type' ), '' ) AS description
    FROM
        pg_catalog.pg_type t
        JOIN pg_catalog.pg_namespace n
            ON n.oid = t.typnamespace
    WHERE ( t.typrelid = 0
            OR ( SELECT c.relkind = 'c'
                    FROM pg_catalog.pg_class c
                    WHERE c.oid = t.typrelid ) )
        AND NOT EXISTS (
                SELECT 1
                    FROM pg_catalog.pg_type el
                    WHERE el.oid = t.typelem
                    AND el.typarray = t.oid )
        AND n.nspname <> 'pg_catalog'
        AND n.nspname <> 'information_schema'
        AND n.nspname !~ '^pg_toast'
),
cols AS (
    SELECT n.nspname::text AS schema_name,
            /*
            When the schema is public, pg_catalog.format_type does not
            include a schema prefix, but when its any other schema, it does.
            This function removes the schema prefix if it exists for
            consistency
            */
            (
                string_to_array(
                    pg_catalog.format_type ( t.oid, NULL ),
                    '.'
                )
            )[array_upper(
                string_to_array(
                    pg_catalog.format_type ( t.oid, NULL ),
                    '.'
                ),
                1)
            ] as obj_name,
            a.attname::text AS column_name,
            pg_catalog.format_type ( a.atttypid, a.atttypmod ) AS data_type,
            a.attnotnull AS is_required,
            a.attnum AS ordinal_position,
            pg_catalog.col_description ( a.attrelid, a.attnum ) AS description
        FROM pg_catalog.pg_attribute a
        JOIN pg_catalog.pg_type t
            ON a.attrelid = t.typrelid
        JOIN pg_catalog.pg_namespace n
            ON ( n.oid = t.typnamespace )
        JOIN types
            ON ( types.nspname = n.nspname
                AND types.obj_name = pg_catalog.format_type ( t.oid, NULL ) )
        WHERE a.attnum > 0
            AND NOT a.attisdropped
)
SELECT
    cols.schema_name,
    cols.obj_name,
    cols.column_name,
    cols.data_type,
    cols.ordinal_position,
    cols.is_required,
    coalesce ( cols.description, '' ) AS description
    FROM
        cols
    WHERE
        cols.schema_name = :schema
    ORDER BY
        cols.schema_name,
        cols.obj_name,
        cols.ordinal_position
    """
    )
    rows = engine.execute(sql, schema=schema).fetchall()

    composites = {}

    for _, composite_rows in flu(rows).group_by(lambda x: x[0] + "." + x[1]):
        # attributed for python class
        attrs = []
        # columns for sqla composite
        columns = []
        for (schema_name, composite_name, column_name, data_type, _, is_required, desc) in composite_rows:  # ordinal
            attrs.append(column_name)
            column_type = type_map.get(data_type, sqltypes.NULLTYPE)
            nullable = not is_required
            column = Column(name=column_name, key=column_name, type_=column_type, nullable=nullable, comment=desc)
            columns.append(column)

        py_composite_name = snake_to_camel(composite_name, upper=True)
        # sqla_composite = CompositeType(py_composite_name, columns)
        sqla_composite = composite_type_factory(py_composite_name, columns, composite_name, schema_name)
        composites[(schema_name, composite_name)] = sqla_composite
    return composites



from __future__ import annotations

from collections import namedtuple
from typing import List, NamedTuple, Type

from sqlalchemy import Column
from sqlalchemy.sql.type_api import TypeEngine
from sqlalchemy.types import UserDefinedType


class CompositeType(UserDefinedType):  # type: ignore
    """
    Represents a PostgreSQL composite type.

    :param name:
        Name of the composite type.
    :param columns:
        List of columns that this composite type consists of
    """

    python_type = tuple

    name: str
    columns: List[Column] = []
    type_cls: Type[NamedTuple]

    pg_name: str
    pg_schema: str

    def init(self, *args, **kwargs):
        pass


def composite_type_factory(name: str, columns: List[Column], pg_name: str, pg_schema: str) -> TypeEngine:

    for column in columns:
        column.key = column.name

    type_cls: Type[NamedTuple] = namedtuple(name, [c.name for c in columns])  # type: ignore

    composite = type(
        name,
        (CompositeType,),
        {
            "name": name,
            "columns": columns,
            "type_cls": type_cls,
            "pg_name": pg_name,
            "pg_schema": pg_schema,
        },  # type:ignore
    )
    return composite  # type: ignore



# pylint: disable=unsubscriptable-object, invalid-name
from __future__ import annotations

from functools import lru_cache
from typing import List, Union

from nebulo.sql.table_base import TableProtocol
from sqlalchemy import Column
from sqlalchemy import inspect as sql_inspect
from sqlalchemy.orm import RelationshipProperty


@lru_cache()
def get_table_name(sqla_model: TableProtocol) -> str:
    """Name of the table"""
    return str(sqla_model.__table__.name)


@lru_cache()
def get_relationships(sqla_model: TableProtocol) -> List[RelationshipProperty]:
    """Relationships with other tables"""
    return list(sql_inspect(sqla_model).relationships)


@lru_cache()
def get_primary_key_columns(sqla_model: TableProtocol) -> List[Column]:
    """Primary key"""
    return [x for x in sqla_model.__table__.primary_key.columns]


@lru_cache()
def get_columns(sqla_model: TableProtocol) -> List[Column]:
    """Columns on the table"""
    return [x for x in sqla_model.__table__.columns]


def get_comment(entity: Union[TableProtocol, Column]) -> str:
    """Get comment on entity"""
    if isinstance(entity, TableProtocol):
        return entity.__table__.comment or ""
    return entity.comment or ""


@lru_cache()
def is_nullable(relationship: RelationshipProperty) -> bool:
    """Checks if a sqlalchemy orm relationship is nullable"""
    for local_col, remote_col in relationship.local_remote_pairs:
        if local_col.nullable or remote_col.nullable:
            return True
    return False



from base64 import b64decode as _unbase64
from base64 import b64encode as _base64

import sqlalchemy
from sqlalchemy import cast, func
from sqlalchemy.dialects.postgresql import BYTEA


def to_base64(string):
    return _base64(string.encode("utf-8")).decode("utf-8")


def from_base64(string):
    return _unbase64(string).decode("utf-8")


def to_base64_sql(text_to_encode):
    encoding = "base64"
    return func.encode(cast(text_to_encode, BYTEA()), cast(encoding, sqlalchemy.Text()))



import os
from pathlib import Path

from appdirs import user_data_dir


class EnvManager:
    """Stashes environment variables in a file and
    retrieves them in (a different process) with get_environ

    with failover to os.environ
    """

    app_env_dir = Path(user_data_dir("NEBULO"))
    app_env = app_env_dir / ".env"

    def __init__(self, **env_vars):
        # Delete if exists
        try:
            os.remove(self.app_env)
        except OSError:
            pass

        self.app_env_dir.mkdir(parents=True, exist_ok=True)
        self.app_env.touch()
        self.vars = env_vars

    def __enter__(self):
        with self.app_env.open("w") as env_file:
            for key, val in self.vars.items():
                env_file.write(f"{key}={val}\n")
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        try:
            os.remove(self.app_env)
        except OSError:
            pass

    @classmethod
    def get_environ(cls):
        try:
            with cls.app_env.open("r") as f:
                for row in f:
                    key, value = row.split("=", 1)
                    os.environ[key.strip()] = value.strip()
        except FileNotFoundError:
            pass
        return os.environ



from nebulo.gql.alias import Boolean, Field, NonNull, ObjectType
from nebulo.gql.relay.cursor import Cursor

PageInfo = ObjectType(
    name="PageInfo",
    fields={
        "hasNextPage": Field(NonNull(Boolean)),
        "hasPreviousPage": Field(NonNull(Boolean)),
        "startCursor": Field(Cursor),
        "endCursor": Field(Cursor),
    },
    description="",
)



from setuptools import find_packages, setup

setup(
    name="nebulo",
    version="0.1.6",
    description="Nebulo: Reflect RDBMS to GraphQL API",
    author="Oliver Rice",
    author_email="oliver@oliverrice.com",
    license="MIT",
    url="https://github.com/olirice/nebulo",
    project_urls={
        "Documentation": "https://olirice.github.io/nebulo/",
        "Source Code": "https://github.com/olirice/nebulo",
    },
    keywords="graphql sqlalchemy sql api python",
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    python_requires=">=3.7",
    package_dir={"": "src"},
    packages=find_packages("src"),
    include_package_data=True,
    entry_points={
        "console_scripts": ["nebulo=nebulo.cli:main", "neb=nebulo.cli:main"],
        "pygments.lexers": ["graphqllexer=nebulo.lexer:GraphQLLexer"],
    },
    install_requires=[
        "sqlalchemy==1.3.15",
        "psycopg2-binary==2.8.4",
        "graphql-core==3.1.2",
        "flask==1.1.1",
        "click==7.1.2",
        "aiofiles==0.5.0",
        "typing-extensions",
        "inflect==4.1.0",
        "cachetools==4.0.0",
        "starlette==0.13.2",
        "databases[postgresql]==0.2.6",
        "uvicorn==0.11.7",
        "parse==1.15.0",
        "flupy==1.1.4",
        "pyjwt==1.7.1",
        "appdirs==1.4.3",
    ],
    extras_require={
        "test": ["pytest", "pytest-cov", "requests"],
        "dev": ["pylint", "black", "sqlalchemy-stubs", "pre-commit"],
        "nvim": ["neovim", "python-language-server"],
        "docs": ["mkdocs", "pygments", "pymdown-extensions"],
    },
)



from __future__ import annotations

import typing

from nebulo.config import Config
from nebulo.gql.alias import ObjectType, Schema
from nebulo.gql.convert.connection import connection_field_factory
from nebulo.gql.convert.create import create_entrypoint_factory
from nebulo.gql.convert.delete import delete_entrypoint_factory
from nebulo.gql.convert.function import (
    immutable_function_entrypoint_factory,
    is_jwt_function,
    mutable_function_entrypoint_factory,
)
from nebulo.gql.convert.table import table_field_factory
from nebulo.gql.convert.update import update_entrypoint_factory
from nebulo.gql.resolve.resolvers.asynchronous import async_resolver as resolver
from nebulo.sql.inspect import get_table_name
from nebulo.sql.reflection.function import SQLFunction
from nebulo.text_utils import snake_to_camel, to_plural

__all__ = ["sqla_models_to_graphql_schema"]


def sqla_models_to_graphql_schema(
    sqla_models,
    sql_functions: typing.List[SQLFunction],
    jwt_identifier: typing.Optional[str] = None,
    jwt_secret: typing.Optional[str] = None,
) -> Schema:
    """Creates a GraphQL Schema from SQLA Models"""

    query_fields = {}
    mutation_fields = {}

    # Tables
    for sqla_model in sqla_models:

        if not Config.exclude_read(sqla_model):
            # e.g. account(nodeId: NodeID)
            single_name = snake_to_camel(get_table_name(sqla_model), upper=False)
            query_fields[single_name] = table_field_factory(sqla_model, resolver)

            # e.g. allAccounts(first: Int, last: Int ....)
            connection_name = "all" + snake_to_camel(to_plural(get_table_name(sqla_model)), upper=True)
            query_fields[connection_name] = connection_field_factory(sqla_model, resolver)

        if not Config.exclude_insert(sqla_model):
            # e.g. createAccount(input: CreateAccountInput)
            mutation_fields.update(create_entrypoint_factory(sqla_model, resolver=resolver))

        if not Config.exclude_update(sqla_model):
            # e.g. updateAccount(input: UpdateAccountInput)
            mutation_fields.update(update_entrypoint_factory(sqla_model, resolver=resolver))

        if not Config.exclude_delete(sqla_model):
            # e.g. deleteAccount(input: DeleteAccountInput)
            mutation_fields.update(delete_entrypoint_factory(sqla_model, resolver=resolver))

    # Functions
    for sql_function in sql_functions:
        if is_jwt_function(sql_function, jwt_identifier):
            mutation_fields.update(
                mutable_function_entrypoint_factory(sql_function=sql_function, resolver=resolver, jwt_secret=jwt_secret)
            )
        else:

            # Immutable functions are queries
            if sql_function.is_immutable:
                query_fields.update(immutable_function_entrypoint_factory(sql_function=sql_function, resolver=resolver))

            # Mutable functions are mutations
            else:
                mutation_fields.update(
                    mutable_function_entrypoint_factory(sql_function=sql_function, resolver=resolver)
                )

    schema_kwargs = {
        "query": ObjectType(name="Query", fields=query_fields),
        "mutation": ObjectType(name="Mutation", fields=mutation_fields),
    }
    return Schema(**{k: v for k, v in schema_kwargs.items() if v.fields})



from __future__ import annotations

from functools import lru_cache

from nebulo.config import Config
from nebulo.gql.alias import Argument, ConnectionType, EdgeType, Field, InputObjectType, Int, List, NonNull
from nebulo.gql.convert.column import convert_column_to_input
from nebulo.gql.relay.cursor import Cursor
from nebulo.gql.relay.page_info import PageInfo
from nebulo.gql.resolve.resolvers.default import default_resolver
from nebulo.sql.inspect import get_columns
from nebulo.sql.table_base import TableProtocol

__all__ = ["connection_field_factory"]


@lru_cache()
def connection_field_factory(sqla_model: TableProtocol, resolver, not_null=False) -> Field:
    relevant_type_name = Config.table_type_name_mapper(sqla_model)
    connection = connection_factory(sqla_model)
    condition = condition_factory(sqla_model)
    args = {
        "first": Argument(Int, description="", out_name=None),
        "last": Argument(Int),
        "before": Argument(Cursor),
        "after": Argument(Cursor),
        "condition": Argument(condition),
    }
    return Field(
        NonNull(connection) if not_null else connection,
        args=args,
        resolve=resolver,
        description=f"Reads and enables pagination through a set of {relevant_type_name}",
    )


@lru_cache()
def edge_factory(sqla_model: TableProtocol) -> EdgeType:
    from .table import table_factory

    name = Config.table_type_name_mapper(sqla_model) + "Edge"

    def build_attrs():
        return {"cursor": Field(Cursor), "node": Field(table_factory(sqla_model))}

    edge = EdgeType(name=name, fields=build_attrs, description="", sqla_model=sqla_model)
    return edge


@lru_cache()
def condition_factory(sqla_model: TableProtocol) -> InputObjectType:
    result_name = f"{Config.table_name_mapper(sqla_model)}Condition"

    attrs = {}
    for column in get_columns(sqla_model):
        field_key = Config.column_name_mapper(column)
        attrs[field_key] = convert_column_to_input(column)
    return InputObjectType(result_name, attrs, description="")


@lru_cache()
def connection_factory(sqla_model: TableProtocol) -> ConnectionType:
    name = Config.table_type_name_mapper(sqla_model) + "Connection"

    def build_attrs():
        edge = edge_factory(sqla_model)
        return {
            "edges": Field(NonNull(List(NonNull(edge))), resolve=default_resolver),
            "pageInfo": Field(NonNull(PageInfo), resolve=default_resolver),
            "totalCount": Field(NonNull(Int), resolve=default_resolver),
        }

    return_type = ConnectionType(name=name, fields=build_attrs, description="", sqla_model=sqla_model)
    return return_type



# pylint: disable=missing-class-docstring,invalid-name
import typing

from graphql.language import (
    InputObjectTypeDefinitionNode,
    InputObjectTypeExtensionNode,
    ObjectTypeDefinitionNode,
    ObjectTypeExtensionNode,
)
from graphql.type import (
    GraphQLArgument,
    GraphQLBoolean,
    GraphQLEnumType,
    GraphQLEnumValue,
    GraphQLField,
    GraphQLFieldMap,
    GraphQLFloat,
    GraphQLID,
    GraphQLInputFieldMap,
    GraphQLInputObjectType,
    GraphQLInt,
    GraphQLInterfaceType,
    GraphQLIsTypeOfFn,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLResolveInfo,
    GraphQLScalarType,
    GraphQLSchema,
    GraphQLString,
    GraphQLType,
    Thunk,
)
from graphql.type.definition import GraphQLInputFieldOutType
from nebulo.sql.composite import CompositeType as SQLACompositeType

# Handle name changes from graphql-core and graphql-core-next
try:
    from graphql.type import GraphQLInputObjectField as GraphQLInputField
except ImportError:
    from graphql.type import GraphQLInputField

Type = GraphQLType
List = GraphQLList
NonNull = GraphQLNonNull
Argument = GraphQLArgument
Boolean = GraphQLBoolean
String = GraphQLString
ScalarType = GraphQLScalarType
ID = GraphQLID
InterfaceType = GraphQLInterfaceType
Int = GraphQLInt
InputField = GraphQLInputField
ResolveInfo = GraphQLResolveInfo
EnumType = GraphQLEnumType
EnumValue = GraphQLEnumValue
Schema = GraphQLSchema
Field = GraphQLField
Float = GraphQLFloat


class HasSQLAModel:  # pylint: disable= too-few-public-methods
    sqla_table = None


class HasSQLFunction:  # pylint: disable= too-few-public-methods
    sql_function = None


class HasSQLAComposite:  # pylint: disable= too-few-public-methods
    sqla_composite: SQLACompositeType


class ObjectType(GraphQLObjectType, HasSQLAModel):
    def __init__(
        self,
        name: str,
        fields: Thunk[GraphQLFieldMap],
        interfaces: typing.Optional[Thunk[typing.Collection["GraphQLInterfaceType"]]] = None,
        is_type_of: typing.Optional[GraphQLIsTypeOfFn] = None,
        extensions: typing.Optional[typing.Dict[str, typing.Any]] = None,
        description: typing.Optional[str] = None,
        ast_node: typing.Optional[ObjectTypeDefinitionNode] = None,
        extension_ast_nodes: typing.Optional[typing.Collection[ObjectTypeExtensionNode]] = None,
        sqla_model=None,
    ) -> None:
        super().__init__(
            name=name,
            fields=fields,
            interfaces=interfaces,
            is_type_of=is_type_of,
            extensions=extensions,
            description=description,
            ast_node=ast_node,
            extension_ast_nodes=extension_ast_nodes,
        )
        self.sqla_model = sqla_model


class ConnectionType(ObjectType):
    pass


class EdgeType(ObjectType):
    pass


class TableType(ObjectType):
    pass


class CompositeType(ObjectType, HasSQLAComposite):
    pass


class MutationPayloadType(ObjectType):
    pass


class CreatePayloadType(MutationPayloadType):
    pass


class UpdatePayloadType(MutationPayloadType):
    pass


class DeletePayloadType(MutationPayloadType):
    pass


class FunctionPayloadType(MutationPayloadType, HasSQLFunction):
    pass


class InputObjectType(GraphQLInputObjectType, HasSQLAModel):
    def __init__(
        self,
        name: str,
        fields: Thunk[GraphQLInputFieldMap],
        description: typing.Optional[str] = None,
        out_type: typing.Optional[GraphQLInputFieldOutType] = None,
        extensions: typing.Optional[typing.Dict[str, typing.Any]] = None,
        ast_node: typing.Optional[InputObjectTypeDefinitionNode] = None,
        extension_ast_nodes: typing.Optional[typing.Collection[InputObjectTypeExtensionNode]] = None,
        sqla_model=None,
    ) -> None:
        super().__init__(
            name=name,
            fields=fields,
            description=description,
            out_type=out_type,
            extensions=extensions,
            ast_node=ast_node,
            extension_ast_nodes=extension_ast_nodes,
        )
        self.sqla_model = sqla_model


class CreateInputType(InputObjectType):
    pass


class TableInputType(InputObjectType):
    pass


class UpdateInputType(InputObjectType):
    pass


class DeleteInputType(InputObjectType):
    pass


class FunctionInputType(GraphQLInputObjectType):
    def __init__(
        self,
        name: str,
        fields: Thunk[GraphQLInputFieldMap],
        description: typing.Optional[str] = None,
        out_type: typing.Optional[GraphQLInputFieldOutType] = None,
        extensions: typing.Optional[typing.Dict[str, typing.Any]] = None,
        ast_node: typing.Optional[InputObjectTypeDefinitionNode] = None,
        extension_ast_nodes: typing.Optional[typing.Collection[InputObjectTypeExtensionNode]] = None,
        sql_function=None,
    ) -> None:
        super().__init__(
            name=name,
            fields=fields,
            description=description,
            out_type=out_type,
            extensions=extensions,
            ast_node=ast_node,
            extension_ast_nodes=extension_ast_nodes,
        )
        self.sql_function = sql_function



import re
from functools import lru_cache

__all__ = ["snake_to_camel"]

_re_snake_to_camel = re.compile(r"(_)([a-z\d])")


@lru_cache()
def snake_to_camel(s: str, upper: bool = True) -> str:
    """Convert from snake_case to CamelCase
    If upper is set, then convert to upper CamelCase, otherwise the first character
    keeps its case.
    """
    s = _re_snake_to_camel.sub(lambda m: m.group(2).upper(), s)
    if upper:
        s = s[:1].upper() + s[1:]
    return s
